﻿using System;
using System.IO;
using System.Net.Mail;
using System.Net.Mime;
using System.Text;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Serialization;
using MammothMicro.SMTPGuard;

namespace MammothMicro.SmtpGuardFilter.Common
{
    /// <summary>
    /// Represents a serializable mail message.
    /// </summary>
    public class SerializableMailMessage : IXmlSerializable
    {
        #region Properties

        /// <summary>
        /// Identifier
        /// </summary>
        public Guid Id { get; set; }

        /// <summary>
        /// Main MailMessage object
        /// </summary>
        public MailMessage Email { get; set; }

        /// <summary>
        /// Date and time when this object has been added to database
        /// </summary>
        public DateTime DateAdded { get; set; }

        /// <summary>
        /// Mail from
        /// </summary>
        public string Sender
        {
            get { return this.Email.From.Address; }
        }

        /// <summary>
        /// Default recipient
        /// </summary>
        public string Recipient
        {
            get { return this.Email.To[0].Address; }
        }

        /// <summary>
        ///  Mail subject
        /// </summary>
        public string Subject
        {
            get { return this.Email.Subject; }
        }

        ///<summary>
        /// Filter which blocked message
        ///</summary>
        public Tools.Filters FilteredBy{ get; set;}
        #endregion

        #region Constructors

        /// <summary>
        /// Public consturctor
        /// </summary>
        public SerializableMailMessage()
        {
            Email = new MailMessage();
        }

        /// <summary>
        /// C'tor
        /// </summary>
        /// <param name="email"></param>
        /// <param name="id"></param>
        public SerializableMailMessage(MailMessage email, Guid id, Tools.Filters filteredBy)
        {
            Email = email;
            DateAdded = DateTime.Now;
            Id = id;
        }

        #endregion

        #region IXmlSerializable Members

        /// <summary>
        /// Implementation of GetSchema for serialization
        /// </summary>
        /// <returns>An XML schema</returns>
        public XmlSchema GetSchema()
        {
            return null;
        }

        /// <summary>
        /// Implementation of ReadXml for serialization
        /// </summary>
        /// <param name="reader">XmlReader that contains the data of serialized object</param>
        public void ReadXml(XmlReader reader)
        {
            var xml = new XmlDocument();
            xml.Load(reader);

            // Properties
            XmlNode rootNode = GetConfigSection(xml, "SerializableMailMessage/MailMessage");
            Email.IsBodyHtml = Convert.ToBoolean(rootNode.Attributes["IsBodyHtml"].Value);
            Email.Priority = (MailPriority)Convert.ToInt16(rootNode.Attributes["Priority"].Value);

            // From
            XmlNode fromNode = GetConfigSection(xml, "SerializableMailMessage/MailMessage/From");
            string fromDisplayName = string.Empty;
            if (fromNode.Attributes["DisplayName"] != null)
                fromDisplayName = fromNode.Attributes["DisplayName"].Value;
            var fromAddress = new MailAddress(fromNode.InnerText, fromDisplayName);
            Email.From = fromAddress;

            // To
            XmlNode toNode = GetConfigSection(xml, "SerializableMailMessage/MailMessage/To/Addresses");
            foreach (XmlNode node in toNode.ChildNodes)
            {
                string toDisplayName = string.Empty;
                if (node.Attributes["DisplayName"] != null)
                    toDisplayName = node.Attributes["DisplayName"].Value;
                var toAddress = new MailAddress(node.InnerText, toDisplayName);
                Email.To.Add(toAddress);
            }

            // CC
            XmlNode ccNode = GetConfigSection(xml, "SerializableMailMessage/MailMessage/CC/Addresses");
            if (ccNode != null)
            {
                foreach (XmlNode node in ccNode.ChildNodes)
                {
                    string ccDisplayName = string.Empty;
                    if (node.Attributes["DisplayName"] != null)
                        ccDisplayName = node.Attributes["DisplayName"].Value;
                    var ccAddress = new MailAddress(node.InnerText, ccDisplayName);
                    Email.CC.Add(ccAddress);
                }
            }

            // Bcc
            XmlNode bccNode = GetConfigSection(xml, "SerializableMailMessage/MailMessage/Bcc/Addresses");
            if (bccNode != null)
            {
                foreach (XmlNode node in bccNode.ChildNodes)
                {
                    string bccDisplayName = string.Empty;
                    if (node.Attributes["DisplayName"] != null)
                        bccDisplayName = node.Attributes["DisplayName"].Value;
                    var bccAddress = new MailAddress(node.InnerText, bccDisplayName);
                    Email.Bcc.Add(bccAddress);
                }
            }

            // Subject
            XmlNode subjectNode = GetConfigSection(xml, "SerializableMailMessage/MailMessage/Subject");
            Email.Subject = subjectNode.InnerText;

            // Body
            XmlNode bodyNode = GetConfigSection(xml, "SerializableMailMessage/MailMessage/Body");
            Email.Body = bodyNode.InnerText;

            // ReplyTo
            XmlNode replyToNode = GetConfigSection(xml, "SerializableMailMessage/MailMessage/ReplyTo");
            if (replyToNode != null)
            {
                string replytoDisplayName = string.Empty;
                if (replyToNode.Attributes["DisplayName"] != null)
                    replytoDisplayName = replyToNode.Attributes["DisplayName"].Value;
                var replytoAddress = new MailAddress(replyToNode.InnerText, replytoDisplayName);
                Email.ReplyTo = replytoAddress;
            }

            // Sender
            XmlNode senderNode = GetConfigSection(xml, "SerializableMailMessage/MailMessage/Sender");
            if (senderNode != null)
            {
                string senderDisplayName = string.Empty;
                if (senderNode.Attributes["DisplayName"] != null)
                    senderDisplayName = senderNode.Attributes["DisplayName"].Value;
                var senderAddress = new MailAddress(senderNode.InnerText, senderDisplayName);
                Email.Sender = senderAddress;
            }

            // Attachments
            XmlNode attachmentsNode = GetConfigSection(xml, "SerializableMailMessage/MailMessage/Attachments");
            if (attachmentsNode != null)
            {
                foreach (XmlNode node in attachmentsNode.ChildNodes)
                {
                    string contentTypeString = string.Empty;
                    if (node.Attributes["ContentType"] != null)
                        contentTypeString = node.Attributes["ContentType"].Value;

                    var contentType = new ContentType(contentTypeString);

                    var stream = new MemoryStream();
                    byte[] data = Encoding.UTF8.GetBytes(node.InnerText);
                    stream.Write(data, 0, data.Length);

                    var attachment = new Attachment(stream, contentType);
                    Email.Attachments.Add(attachment);
                }
            }
        }

        /// <summary>
        /// Implementation of WriteXml for serialization
        /// </summary>
        /// <param name="writer">XmlWriter that can be used to serialize the object</param>
        public void WriteXml(XmlWriter writer)
        {
            if (Email != null)
            {
                writer.WriteStartElement("MailMessage");
                writer.WriteAttributeString("Priority", Convert.ToInt16(Email.Priority).ToString());
                writer.WriteAttributeString("IsBodyHtml", Email.IsBodyHtml.ToString());

                // From
                writer.WriteStartElement("From");
                if (!string.IsNullOrEmpty(Email.From.DisplayName))
                    writer.WriteAttributeString("DisplayName", Email.From.DisplayName);
                writer.WriteRaw(Email.From.Address);
                writer.WriteEndElement();

                // To
                writer.WriteStartElement("To");
                writer.WriteStartElement("Addresses");
                foreach (var address in Email.To)
                {
                    writer.WriteStartElement("Address");
                    if (!string.IsNullOrEmpty(address.DisplayName))
                        writer.WriteAttributeString("DisplayName", address.DisplayName);
                    writer.WriteRaw(address.Address);
                    writer.WriteEndElement();
                }
                writer.WriteEndElement();
                writer.WriteEndElement();

                // CC
                if (Email.CC.Count > 0)
                {
                    writer.WriteStartElement("CC");
                    writer.WriteStartElement("Addresses");
                    foreach (var address in Email.CC)
                    {
                        writer.WriteStartElement("Address");
                        if (!string.IsNullOrEmpty(address.DisplayName))
                            writer.WriteAttributeString("DisplayName", address.DisplayName);
                        writer.WriteRaw(address.Address);
                        writer.WriteEndElement();
                    }
                    writer.WriteEndElement();
                    writer.WriteEndElement();
                }

                // Bcc
                if (Email.Bcc.Count > 0)
                {
                    writer.WriteStartElement("Bcc");
                    writer.WriteStartElement("Addresses");
                    foreach (var address in Email.Bcc)
                    {
                        writer.WriteStartElement("Address");
                        if (!string.IsNullOrEmpty(address.DisplayName))
                            writer.WriteAttributeString("DisplayName", address.DisplayName);
                        writer.WriteRaw(address.Address);
                        writer.WriteEndElement();
                    }
                    writer.WriteEndElement();
                    writer.WriteEndElement();
                }

                // Subject
                writer.WriteStartElement("Subject");
                writer.WriteRaw(Email.Subject);
                writer.WriteEndElement();

                // Body
                writer.WriteStartElement("Body");
                writer.WriteCData(Email.Body);
                writer.WriteEndElement();

                // ReplyTo
                if (Email.ReplyTo != null)
                {
                    writer.WriteStartElement("ReplyTo");
                    if (!string.IsNullOrEmpty(Email.ReplyTo.DisplayName))
                        writer.WriteAttributeString("DisplayName", Email.ReplyTo.DisplayName);
                    writer.WriteRaw(Email.ReplyTo.Address);
                    writer.WriteEndElement();
                }

                // Sender
                if (Email.Sender != null)
                {
                    writer.WriteStartElement("Sender");
                    if (!string.IsNullOrEmpty(Email.Sender.DisplayName))
                        writer.WriteAttributeString("DisplayName", Email.Sender.DisplayName);
                    writer.WriteRaw(Email.Sender.Address);
                    writer.WriteEndElement();
                }

                // Attachments
                if (Email.Attachments.Count > 0)
                {
                    writer.WriteStartElement("Attachments");

                    foreach (var attachment in Email.Attachments)
                    {
                        writer.WriteStartElement("Attachment");

                        if (!string.IsNullOrEmpty(attachment.Name))
                            writer.WriteAttributeString("ContentType", attachment.ContentType.ToString());

                        using (var reader = new BinaryReader(attachment.ContentStream))
                        {
                            byte[] data = reader.ReadBytes((int)attachment.ContentStream.Length);

                            writer.WriteBase64(data, 0, data.Length);
                        }

                        writer.WriteEndElement();
                    }

                    writer.WriteEndElement();
                }

                writer.WriteEndElement();
            }
        }

        #endregion

        #region Private Methods

        /// <summary>
        /// REturns a specific node in the serialized XML
        /// </summary>
        /// <param name="xml">XML document</param>
        /// <param name="nodePath">XPath expression for the node path</param>
        /// <returns>An XML node</returns>
        private XmlNode GetConfigSection(XmlDocument xml, string nodePath)
        {
            return xml.SelectSingleNode(nodePath);
        }

        #endregion
    }
}


